# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export the spans data to python logging."""

import logging

from opencensus.common.transports import sync
from opencensus.trace import base_exporter, span_data


class LoggingExporter(base_exporter.Exporter):
    """A exporter to export the spans data to python logging. Also can use
    handlers like CloudLoggingHandler to log to Stackdriver Logging API.

    :type handler: :class:`logging.handler`
    :param handler: the handler to attach to the global handler

    :type transport: :class:`type`
    :param transport: Class for creating new transport objects. It should
                      extend from the base_exporter :class:`.Transport` type
                      and implement :meth:`.Transport.export`. Defaults to
                      :class:`.SyncTransport`. The other option is
                      :class:`.AsyncTransport`.

    Example:

    .. code-block:: python

        import google.cloud.logging
        from google.cloud.logging.handlers import CloudLoggingHandler
        from opencensus.trace import logging_exporter

        client = google.cloud.logging.Client()
        cloud_handler = CloudLoggingHandler(client)
        exporter = logging_exporter.LoggingExporter(handler=cloud_handler)

        exporter.export(your_spans_list)

    Or initialize a context tracer with the logging exporter, then the traces
    will be exported to logging when finished.
    """

    def __init__(self, handler=None, transport=sync.SyncTransport):
        self.logger = logging.getLogger()

        if handler is None:
            handler = logging.StreamHandler()

        self.handler = handler
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        self.transport = transport(self)

    def emit(self, span_datas):
        """
        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to emit
        """
        # convert to the legacy trace json for easier refactoring
        # TODO: refactor this to use the span data directly
        legacy_trace_json = span_data.format_legacy_trace_json(span_datas)
        self.logger.info(legacy_trace_json)

    def export(self, span_datas):
        """
        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to export
        """
        self.transport.export(span_datas)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
from collections import defaultdict

from opencensus.stats import metric_utils
from opencensus.stats import view_data as view_data_module

logger = logging.getLogger(__name__)


class MeasureToViewMap(object):
    """Measure To View Map stores a map from names of Measures to
    specific View Datas

    """

    def __init__(self):
        # stores the one-to-many mapping from Measures to View Datas
        self._measure_to_view_data_list_map = defaultdict(list)
        # stores a map from the registered View names to the Views
        self._registered_views = {}
        # stores a map from the registered Measure names to the Measures
        self._registered_measures = {}
        # stores the set of the exported views
        self._exported_views = set()
        # Stores the registered exporters
        self._exporters = []

    @property
    def exported_views(self):
        """the current exported views"""
        return self._exported_views

    @property
    def exporters(self):
        """registered exporters"""
        return self._exporters

    def get_view(self, view_name, timestamp):
        """get the View Data from the given View name"""
        view = self._registered_views.get(view_name)
        if view is None:
            return None

        view_data_list = self._measure_to_view_data_list_map.get(
            view.measure.name)

        if not view_data_list:
            return None

        for view_data in view_data_list:
            if view_data.view.name == view_name:
                break
        else:
            return None

        return self.copy_and_finalize_view_data(view_data)

    def filter_exported_views(self, all_views):
        """returns the subset of the given view that should be exported"""
        views = set(all_views)
        return views

    # TODO: deprecate
    def register_view(self, view, timestamp):
        """registers the view's measure name to View Datas given a view"""
        if len(self.exporters) > 0:
            try:
                for e in self.exporters:
                    e.on_register_view(view)
            except AttributeError:
                pass

        self._exported_views = None
        existing_view = self._registered_views.get(view.name)
        if existing_view is not None:
            if existing_view == view:
                # ignore the views that are already registered
                return
            else:
                logger.warning(
                    "A different view with the same name is already registered"
                )  # pragma: NO COVER
        measure = view.measure
        registered_measure = self._registered_measures.get(measure.name)
        if registered_measure is not None and registered_measure != measure:
            logger.warning(
                "A different measure with the same name is already registered")
        self._registered_views[view.name] = view
        if registered_measure is None:
            self._registered_measures[measure.name] = measure
        self._measure_to_view_data_list_map[view.measure.name].append(
            view_data_module.ViewData(view=view, start_time=timestamp,
                                      end_time=timestamp))

    def record(self, tags, measurement_map, timestamp, attachments=None):
        """records stats with a set of tags"""
        assert all(vv >= 0 for vv in measurement_map.values())
        for measure, value in measurement_map.items():
            if measure != self._registered_measures.get(measure.name):
                return
            view_datas = []
            for measure_name, view_data_list \
                    in self._measure_to_view_data_list_map.items():
                if measure_name == measure.name:
                    view_datas.extend(view_data_list)
            for view_data in view_datas:
                view_data.record(
                    context=tags, value=value, timestamp=timestamp,
                    attachments=attachments)
            self.export(view_datas)

    # TODO: deprecate
    def export(self, view_datas):
        """export view datas to registered exporters"""
        view_datas_copy = \
            [self.copy_and_finalize_view_data(vd) for vd in view_datas]
        if len(self.exporters) > 0:
            for e in self.exporters:
                try:
                    e.export(view_datas_copy)
                except AttributeError:
                    pass

    def get_metrics(self, timestamp):
        """Get a Metric for each registered view.

        Convert each registered view's associated `ViewData` into a `Metric` to
        be exported.

        :type timestamp: :class: `datetime.datetime`
        :param timestamp: The timestamp to use for metric conversions, usually
        the current time.

        :rtype: Iterator[:class: `opencensus.metrics.export.metric.Metric`]
        """
        for vdl in self._measure_to_view_data_list_map.values():
            for vd in vdl:
                metric = metric_utils.view_data_to_metric(vd, timestamp)
                if metric is not None:
                    yield metric

    # TODO(issue #470): remove this method once we export immutable stats.
    def copy_and_finalize_view_data(self, view_data):
        view_data_copy = copy.copy(view_data)
        tvdam_copy = copy.deepcopy(view_data.tag_value_aggregation_data_map)
        view_data_copy._tag_value_aggregation_data_map = tvdam_copy
        view_data_copy.end()
        return view_data_copy



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class TimeSeries(object):
    """Time series data for a given metric and time interval.

    This class implements the spec for v1 TimeSeries structs as of
    opencensus-proto release v0.1.0. See opencensus-proto for details:

        https://github.com/census-instrumentation/opencensus-proto/blob/v0.1.0/src/opencensus/proto/metrics/v1/metrics.proto#L132

    A TimeSeries is a collection of data points that describes the time-varying
    values of a metric.

    :type label_values: list(:class:
    '~opencensus.metrics.label_value.LabelValue')
    :param label_values: The set of label values that uniquely identify this
    timeseries.

    :type points: list(:class: '~opencensus.metrics.export.point.Point')
    :param points: The data points of this timeseries.

    :type start_timestamp: str
    :param start_timestamp: The time when the cumulative value was reset to
    zero, must be set for cumulative metrics.
    """  # noqa

    def __init__(self, label_values, points, start_timestamp):
        if label_values is None:
            raise ValueError("label_values must not be None")
        if not points:
            raise ValueError("points must not be null or empty")
        self._label_values = label_values
        self._points = points
        self._start_timestamp = start_timestamp

    def __repr__(self):
        points_repr = '[{}]'.format(
            ', '.join(repr(point.value) for point in self.points))

        lv_repr = tuple(lv.value for lv in self.label_values)
        return ('{}({}, label_values={}, start_timestamp={})'
                .format(
                    type(self).__name__,
                    points_repr,
                    lv_repr,
                    self.start_timestamp
                ))

    @property
    def start_timestamp(self):
        return self._start_timestamp

    @property
    def label_values(self):
        return self._label_values

    @property
    def points(self):
        return self._points

    def check_points_type(self, type_class):
        """Check that each point's value is an instance of `type_class`.

        `type_class` should typically be a Value type, i.e. one that extends
        :class: `opencensus.metrics.export.value.Value`.

        :type type_class: type
        :param type_class: Type to check against.

        :rtype: bool
        :return: Whether all points are instances of `type_class`.
        """
        for point in self.points:
            if (point.value is not None
                    and not isinstance(point.value, type_class)):
                return False
        return True



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from opencensus.common.http_handler import get_request

REGION_KEY = 'region'
ACCOUNT_ID_KEY = 'aws_account'
INSTANCE_ID_KEY = 'instance_id'

# AWS provides Instance Metadata via below url
_AWS_INSTANCE_IDENTITY_DOCUMENT_URI = \
    "http://169.254.169.254/latest/dynamic/instance-identity/document"

_AWS_ATTRIBUTES = {
    # Region is the AWS region for the VM. The format of this field is
    # "aws:{region}", where supported values for {region} are listed at
    # http://docs.aws.amazon.com/general/latest/gr/rande.html.
    'region': REGION_KEY,

    # accountId is the AWS account number for the VM.
    'accountId': ACCOUNT_ID_KEY,

    # instanceId is the instance id of the instance.
    'instanceId': INSTANCE_ID_KEY
}

# inited is used to make sure AWS initialize executes only once.
inited = False

# Detects if the application is running on EC2 by making a connection to AWS
# instance identity document URI.If connection is successful, application
# should be on an EC2 instance.
is_running_on_aws = False

aws_metadata_map = {}


class AwsIdentityDocumentUtils(object):
    """Util methods for getting and parsing AWS instance identity document."""

    inited = False
    is_running = False

    @classmethod
    def _initialize_aws_identity_document(cls):
        """This method, tries to establish an HTTP connection to AWS instance
        identity document url. If the application is running on an EC2
        instance, we should be able to get back a valid JSON document. Make a
        http get request call and store data in local map.
        This method should only be called once.
        """

        if cls.inited:
            return

        content = get_request(_AWS_INSTANCE_IDENTITY_DOCUMENT_URI)
        if content is not None:
            content = json.loads(content)
            for env_var, attribute_key in _AWS_ATTRIBUTES.items():
                attribute_value = content.get(env_var)
                if attribute_value is not None:
                    aws_metadata_map[attribute_key] = attribute_value

            cls.is_running = True

        cls.inited = True

    @classmethod
    def is_running_on_aws(cls):
        cls._initialize_aws_identity_document()
        return cls.is_running

    def get_aws_metadata(self):
        """AWS Instance Identity Document is a JSON file.
        See docs.aws.amazon.com/AWSEC2/latest/UserGuide/
        instance-identity-documents.html.
        :return:
        """
        if self.is_running_on_aws():
            return aws_metadata_map

        return dict()



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import collections
import logging
import struct

from opencensus.trace import span_context as span_context_module
from opencensus.trace.trace_options import TraceOptions

# Used for decoding hex bytes to hex string.
UTF8 = 'utf-8'

VERSION_ID = 0
TRACE_ID_FIELD_ID = 0
SPAN_ID_FIELD_ID = 1
TRACE_OPTION_FIELD_ID = 2

# Sizes are number of bytes.
ID_SIZE = 1
TRACE_ID_SIZE = 16
SPAN_ID_SIZE = 8
TRACE_OPTION_SIZE = 1

FORMAT_LENGTH = 4 * ID_SIZE + TRACE_ID_SIZE + SPAN_ID_SIZE + TRACE_OPTION_SIZE

# See: https://docs.python.org/3/library/struct.html#format-characters
BIG_ENDIAN = '>'
CHAR_ARRAY_FORMAT = 's'
UNSIGNED_CHAR = 'B'
UNSIGNED_LONG_LONG = 'Q'

# Adding big endian indicator at the beginning to avoid auto padding. This is
# for ensuring the length of binary is not changed when propagating.
BINARY_FORMAT = '{big_endian}{version_id}' \
    '{trace_id_field_id}{trace_id}' \
    '{span_id_field_id}{span_id}' \
    '{trace_option_field_id}{trace_option}'\
    .format(
        big_endian=BIG_ENDIAN,
        version_id=UNSIGNED_CHAR,
        trace_id_field_id=UNSIGNED_CHAR,
        trace_id='{}{}'.format(TRACE_ID_SIZE, CHAR_ARRAY_FORMAT),
        span_id_field_id=UNSIGNED_CHAR,
        span_id='{}{}'.format(SPAN_ID_SIZE, CHAR_ARRAY_FORMAT),
        trace_option_field_id=UNSIGNED_CHAR,
        trace_option=UNSIGNED_CHAR)

Header = collections.namedtuple(
    'Header',
    'version_id '
    'trace_id_field_id '
    'trace_id '
    'span_id_field_id '
    'span_id '
    'trace_option_field_id '
    'trace_option')


class BinaryFormatPropagator(object):
    """This propagator contains the method for serializing and deserializing
    SpanContext using a binary format.

    See: https://github.com/census-instrumentation/opencensus-specs/blob/
         master/encodings/BinaryEncoding.md

    Example:
        [SpanContext]
            trace_id: hex string with length 32.
                e.g. 'a0b72ca15c1a4bd18962d0ac59dc90b9'
            span_id: hex string with length 16.
                e.g. 'a0b72ca15c1a4bd1'
            enabled (trace option): bool.
                e.g. True
        [Binary Format]
            trace_id: Bytes with length 16.
                e.g. b'\xa0\xb7,\xa1\\\x1aK\xd1\x89b\xd0\xacY\xdc\x90\xb9'
            span_id: Bytes with length 8.
                e.g. b'\x00\xf0g\xaa\x0b\xa9\x02\xb7'
            trace_option: Byte with length 1.
                e.g. b'\x01'
    """
    def from_header(self, binary):
        """Generate a SpanContext object using the trace context header.
        The value of enabled parsed from header is int. Need to convert to
        bool.

        :type binary: bytes
        :param binary: Trace context header which was extracted from the
                       request headers.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from the trace context header.
        """
        # If no binary provided, generate a new SpanContext
        if binary is None:
            return span_context_module.SpanContext(from_header=False)

        # If cannot parse, return a new SpanContext and ignore the context
        # from binary.
        try:
            data = Header._make(struct.unpack(BINARY_FORMAT, binary))
        except struct.error:
            logging.warning(
                'Cannot parse the incoming binary data {}, '
                'wrong format. Total bytes length should be {}.'.format(
                    binary, FORMAT_LENGTH
                )
            )
            return span_context_module.SpanContext(from_header=False)

        # data.trace_id is in bytes with length 16, hexlify it to hex bytes
        # with length 32, then decode it to hex string using utf-8.
        trace_id = str(binascii.hexlify(data.trace_id).decode(UTF8))
        span_id = str(binascii.hexlify(data.span_id).decode(UTF8))
        trace_options = TraceOptions(data.trace_option)

        span_context = span_context_module.SpanContext(
                trace_id=trace_id,
                span_id=span_id,
                trace_options=trace_options,
                from_header=True)

        return span_context

    def to_header(self, span_context):
        """Convert a SpanContext object to header in binary format.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :rtype: bytes
        :returns: A trace context header in binary format.
        """
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        trace_options = int(span_context.trace_options.trace_options_byte)

        # If there is no span_id in this context, set it to 0, which is
        # considered invalid and won't be set as the downstream parent span_id.
        if span_id is None:
            span_id = span_context_module.INVALID_SPAN_ID

        # Convert trace_id to bytes with length 16, treat span_id as 64 bit
        # integer which is unsigned long long type and convert it to bytes with
        # length 8, trace_option is integer with length 1.
        return struct.pack(
            BINARY_FORMAT,
            VERSION_ID,
            TRACE_ID_FIELD_ID,
            binascii.unhexlify(trace_id),
            SPAN_ID_FIELD_ID,
            binascii.unhexlify(span_id),
            TRACE_OPTION_FIELD_ID,
            trace_options)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

# Enabled field is the least significant bit of trace options.
_ENABLED_BITMASK = 1 << 0

# Default trace options
DEFAULT = '1'


class TraceOptions(object):
    """A class that represents global trace options.

    :type trace_options_byte: str
    :param trace_options_byte: 1 byte bitmap for trace options.
    """

    def __init__(self, trace_options_byte=None):
        if trace_options_byte is None:
            trace_options_byte = DEFAULT

        self.trace_options_byte = self.check_trace_options(trace_options_byte)
        self.enabled = self.get_enabled()

    def check_trace_options(self, trace_options_byte):
        trace_options_int = int(trace_options_byte)

        if trace_options_int < 0 or trace_options_int > 255:
            logging.warning("Trace options invalid, should be 1 byte.")
            trace_options_byte = DEFAULT

        return trace_options_byte

    def __repr__(self):
        fmt = '{}(enabled={})'
        return fmt.format(
            type(self).__name__,
            self.get_enabled(),
        )

    def get_enabled(self):
        """Get the last bit from the trace options which is the enabled field.

        :type trace_options: byte
        :param trace_options: 1 byte field which indicates 8 trace options,
                              currently only have the enabled option. 1 means
                              enabled, 0 means not enabled.

        :rtype: bool
        :returns: Enabled tracing or not.
        """
        enabled = bool(int(self.trace_options_byte) & _ENABLED_BITMASK)

        return enabled

    def set_enabled(self, enabled):
        """Update the last bit of the trace options byte str.

        :type enabled: bool
        :param enabled: Whether enable tracing in this span context or not.
        """
        enabled_bit = '1' if enabled else '0'
        self.trace_options_byte = str(
            self.trace_options_byte)[:-1] + enabled_bit
        self.enabled = self.get_enabled()



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib

__all__ = ['Namespace', 'load']


class Namespace(object):
    def __init__(self, name, parent=None):
        self.parent = parent
        self.name = name

    def __getattr__(self, name):
        return type(self)(name, self)

    def __str__(self):
        if self.parent is None:
            return self.name
        return '{!s}.{}'.format(self.parent, self.name)

    def __call__(self, *args, **kwargs):
        ctor = getattr(importlib.import_module(str(self.parent)), self.name)
        return ctor(*args, **kwargs)

    @classmethod
    def eval(cls, expr):
        return eval(expr, {}, {'opencensus': cls('opencensus')})


def load(expr):
    """Dynamically import OpenCensus components and evaluate the provided
    configuration expression.
    """
    return Namespace.eval(expr)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Point(object):
    """A timestamped measurement of a TimeSeries.

    :type value: :class:`opencensus.metrics.export.value.ValueDouble` or
        :class:`opencensus.metrics.export.value.ValueLong` or
        :class:`opencensus.metrics.export.value.ValueSummary` or
        :class:`opencensus.metrics.export.value.ValueDistribution`
    :param value: the point value.

    :type timestamp: time
    :param timestamp: the timestamp when the `Point` was recorded.
    """

    def __init__(self, value, timestamp):
        self._value = value
        self._timestamp = timestamp

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    def __repr__(self):
        return ("{}(value={}, timestamp={})"
                .format(
                    type(self).__name__,
                    self.value,
                    self.timestamp
                ))



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-

import six

import logging

from google.protobuf.internal.encoder import _VarintBytes

from opencensus.tags import tag_map as tag_map_module

# Used for decoding hex bytes to hex string.
UTF8 = 'utf-8'

VERSION_ID = 0
TAG_FIELD_ID = 0
TAG_MAP_SERIALIZED_SIZE_LIMIT = 8192


class BinarySerializer(object):
    def from_byte_array(self, binary):
        if len(binary) <= 0:
            logging.warning("Input byte[] cannot be empty/")
            return tag_map_module.TagMap()
        else:
            buffer = memoryview(binary)
            version_id = buffer[0]
            if six.PY2:
                version_id = ord(version_id)
            if version_id != VERSION_ID:
                raise ValueError("Invalid version id.")
            return self._parse_tags(buffer)

    def to_byte_array(self, tag_context):
        encoded_bytes = b''
        encoded_bytes += _VarintBytes(VERSION_ID)
        total_chars = 0
        for tag in tag_context:
            tag_key, tag_value = tag
            total_chars += len(tag_key)
            total_chars += len(tag_value)
            encoded_bytes = self._encode_tag(
                tag_key, tag_value, encoded_bytes)
        if total_chars <= TAG_MAP_SERIALIZED_SIZE_LIMIT:
            return encoded_bytes
        else:  # pragma: NO COVER
            logging.warning("Size of the tag context exceeds the maximum size")

    def _parse_tags(self, buffer):
        tag_context = tag_map_module.TagMap()
        limit = len(buffer)
        total_chars = 0
        i = 1
        while i < limit:
            field_id = buffer[i] if six.PY3 else ord(buffer[i])
            if field_id == TAG_FIELD_ID:
                i += 1
                key = self._decode_string(buffer, i)
                i += len(key)
                total_chars += len(key)
                i += 1
                val = self._decode_string(buffer, i)
                i += len(val)
                total_chars += len(val)
                i += 1
                if total_chars > \
                        TAG_MAP_SERIALIZED_SIZE_LIMIT:  # pragma: NO COVER
                    logging.warning("Size of the tag context exceeds maximum")
                    break
                else:
                    tag_context.insert(str(key), str(val))
            else:
                break
        return tag_context

    def _encode_tag(self, tag_key, tag_value, encoded_bytes):
        encoded_bytes += _VarintBytes(TAG_FIELD_ID)
        encoded_bytes = self._encode_string(tag_key, encoded_bytes)
        encoded_bytes = self._encode_string(tag_value, encoded_bytes)
        return encoded_bytes

    def _encode_string(self, input_str, encoded_bytes):
        encoded_bytes += _VarintBytes(len(input_str))
        encoded_bytes += input_str.encode(UTF8)
        return encoded_bytes

    def _decode_string(self, buffer, pos):
        length = buffer[pos] if six.PY3 else ord(buffer[pos])
        builder = ""
        i = 1
        while i <= length:
            bytes_to_decode = buffer[pos + i] if six.PY3 \
                else ord(buffer[pos + i])
            builder += _VarintBytes(bytes_to_decode).decode()
            i += 1
        return builder



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

COMMON_ATTRIBUTES = {
    'AGENT': 'g.co/agent',
    'COMPONENT': 'component',
    'ERROR_MESSAGE': 'error.message',
    'ERROR_NAME': 'error.name',
    'HTTP_CLIENT_CITY': 'http.client_city',
    'HTTP_CLIENT_COUNTRY': 'http.client_country',
    'HTTP_CLIENT_PROTOCOL': 'http.client_protocol',
    'HTTP_CLIENT_REGION': 'http.client_region',
    'HTTP_HOST': 'http.host',
    'HTTP_METHOD': 'http.method',
    'HTTP_PATH': 'http.path',
    'HTTP_ROUTE': 'http.route',
    'HTTP_REDIRECTED_URL': 'http.redirected_url',
    'HTTP_REQUEST_SIZE': 'http.request_size',
    'HTTP_RESPONSE_SIZE': 'http.response_size',
    'HTTP_STATUS_CODE': 'http.status_code',
    'HTTP_URL': 'http.url',
    'HTTP_USER_AGENT': 'http.user_agent',
    'PID': 'pid',
    'STACKTRACE': 'stacktrace',
    'TID': 'tid',
}


GRPC_ATTRIBUTES = {
    'GRPC_HOST_PORT': 'grpc.host_port',
    'GRPC_METHOD': 'grpc.method',
}



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from opencensus.metrics.export import value


class _MetricDescriptorTypeMeta(type):
    """Helper for `x in MetricDescriptorType`."""

    def __contains__(cls, item):
        return item in {
            MetricDescriptorType.GAUGE_INT64,
            MetricDescriptorType.GAUGE_DOUBLE,
            MetricDescriptorType.GAUGE_DISTRIBUTION,
            MetricDescriptorType.CUMULATIVE_INT64,
            MetricDescriptorType.CUMULATIVE_DOUBLE,
            MetricDescriptorType.CUMULATIVE_DISTRIBUTION
        }


@six.add_metaclass(_MetricDescriptorTypeMeta)
class MetricDescriptorType(object):
    """The kind of metric. It describes how the data is reported.

    MetricDescriptorType is an enum of valid MetricDescriptor type values. See
    opencensus-proto for details:

    https://github.com/census-instrumentation/opencensus-proto/blob/v0.1.0/src/opencensus/proto/metrics/v1/metrics.proto#L79

    A gauge is an instantaneous measurement of a value.

    A cumulative measurement is a value accumulated over a time interval. In a
    time series, cumulative measurements should have the same start time and
    increasing end times, until an event resets the cumulative value to zero
    and sets a new start time for the following points.

    """
    # Integer gauge. The value can go both up and down.
    GAUGE_INT64 = 1

    # Floating point gauge. The value can go both up and down.
    GAUGE_DOUBLE = 2

    # Distribution gauge measurement. The count and sum can go both up and
    # down. Recorded values are always >= 0.
    # Used in scenarios like a snapshot of time the current items in a queue
    # have spent there.
    GAUGE_DISTRIBUTION = 3

    # Integer cumulative measurement. The value cannot decrease, if resets then
    # the start_time should also be reset.
    CUMULATIVE_INT64 = 4

    # Floating point cumulative measurement. The value cannot decrease, if
    # resets then the start_time should also be reset. Recorded values are
    # always >= 0.
    CUMULATIVE_DOUBLE = 5

    # Distribution cumulative measurement. The count and sum cannot decrease,
    # if resets then the start_time should also be reset.
    CUMULATIVE_DISTRIBUTION = 6

    # Some frameworks implemented Histograms as a summary of observations
    # (usually things like request durations and response sizes). While it also
    # provides a total count of observations and a sum of all observed values,
    # it calculates configurable percentiles over a sliding time window. This
    # is not recommended, since it cannot be aggregated.
    SUMMARY = 7

    _type_map = {
        GAUGE_INT64: value.ValueLong,
        GAUGE_DOUBLE: value.ValueDouble,
        GAUGE_DISTRIBUTION: value.ValueDistribution,
        CUMULATIVE_INT64: value.ValueLong,
        CUMULATIVE_DOUBLE: value.ValueDouble,
        CUMULATIVE_DISTRIBUTION: value.ValueDistribution,
        SUMMARY: value.ValueSummary
    }

    @classmethod
    def to_type_class(cls, metric_descriptor_type):
        try:
            return cls._type_map[metric_descriptor_type]
        except KeyError:
            raise ValueError("Unknown MetricDescriptorType value")


class MetricDescriptor(object):
    """Defines a metric type and its schema.

    This class implements the spec for v1 MetricDescriptors, as of
    opencensus-proto release v0.1.0. See opencensus-proto for details:

    https://github.com/census-instrumentation/opencensus-proto/blob/v0.1.0/src/opencensus/proto/metrics/v1/metrics.proto#L59

    :type name: str
    :param name: The metric type, including its DNS name prefix. It must be
    unique.

    :type description: str
    :param description: A detailed description of the metric, which can be used
    in documentation.

    :type unit: str
    :param unit: The unit in which the metric value is reported. Follows the
    format described by http://unitsofmeasure.org/ucum.html.

    :type type_: int
    :param type_: The type of metric. MetricDescriptorType enumerates the valid
    options.

    :type label_keys: list(:class: '~opencensus.metrics.label_key.LabelKey')
    :param label_keys: The label keys associated with the metric descriptor.
    """

    def __init__(self, name, description, unit, type_, label_keys):
        if type_ not in MetricDescriptorType:
            raise ValueError("Invalid type")

        if label_keys is None:
            raise ValueError("label_keys must not be None")

        if any(key is None for key in label_keys):
            raise ValueError("label_keys must not contain null keys")

        self._name = name
        self._description = description
        self._unit = unit
        self._type = type_
        self._label_keys = label_keys

    def __repr__(self):
        type_name = MetricDescriptorType.to_type_class(self.type).__name__
        return ('{}(name="{}", description="{}", unit={}, type={})'
                .format(
                    type(self).__name__,
                    self.name,
                    self.description,
                    self.unit,
                    type_name,
                ))

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def unit(self):
        return self._unit

    @property
    def type(self):
        return self._type

    @property
    def label_keys(self):
        return self._label_keys



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import six

from opencensus.common import utils


def _format_attribute_value(value):
    if isinstance(value, bool):
        value_type = 'bool_value'
    elif isinstance(value, int):
        value_type = 'int_value'
    elif isinstance(value, six.string_types):
        value_type = 'string_value'
        value = utils.get_truncatable_str(value)
    elif isinstance(value, float):
        value_type = 'double_value'
    else:
        return None

    return {value_type: value}


class Attributes(object):
    """A set of attributes, each in the format [KEY]:[VALUE].

    :type attributes: dict
    :param attributes: The set of attributes. Each attribute's key can be up
                       to 128 bytes long. The value can be a string up to 256
                       bytes, an integer, a floating-point number, or the
                       Boolean values true and false.
    """
    def __init__(self, attributes=None):
        self.attributes = attributes or {}

    def set_attribute(self, key, value):
        """Set a key value pair."""
        self.attributes[key] = value

    def delete_attribute(self, key):
        """Delete an attribute given a key if existed."""
        self.attributes.pop(key, None)

    def get_attribute(self, key):
        """Get a attribute value."""
        return self.attributes.get(key, None)

    def format_attributes_json(self):
        """Convert the Attributes object to json format."""
        attributes_json = {}

        for key, value in self.attributes.items():
            key = utils.check_str_length(key)[0]
            value = _format_attribute_value(value)

            if value is not None:
                attributes_json[key] = value

        result = {
            'attributeMap': attributes_json
        }

        return result



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.rpc import code_pb2


class Status(object):
    """The Status type defines a logical error model that is suitable for
    different programming environments, including REST APIs and RPC APIs.
    It is used by gRPC.

    :type code: int
    :param code: An enum value of :class: `~google.rpc.Code`.

    :type message: str
    :param message: A developer-facing error message, should be in English.

    :type details: list
    :param details: A list of messages that carry the error details.
                    There is a common set of message types for APIs to use.
                    e.g. [
                            {
                                "@type": string,
                                field1: ...,
                                ...
                            },
                         ]
                    See: https://cloud.google.com/trace/docs/reference/v2/
                         rest/v2/Status#FIELDS.details
    """
    def __init__(self, code, message=None, details=None):
        self.code = code
        self.message = message
        self.details = details

    @property
    def canonical_code(self):
        return self.code

    @property
    def description(self):
        return self.message

    @property
    def is_ok(self):
        return self.canonical_code == code_pb2.OK

    def format_status_json(self):
        """Convert a Status object to json format."""
        status_json = {}

        status_json['code'] = self.canonical_code

        if self.description is not None:
            status_json['message'] = self.description

        if self.details is not None:
            status_json['details'] = self.details

        return status_json

    @classmethod
    def from_exception(cls, exc):
        return cls(
            code=code_pb2.UNKNOWN,
            message=str(exc)
        )

    @classmethod
    def as_ok(cls):
        return cls(
            code=code_pb2.OK,
        )



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common.runtime_context import RuntimeContext

_measure_to_view_map_slot = RuntimeContext.register_slot(
    'measure_to_view_map',
    lambda: {})


def get_measure_to_view_map():
    return RuntimeContext.measure_to_view_map


def set_measure_to_view_map(measure_to_view_map):
    RuntimeContext.measure_to_view_map = measure_to_view_map


def clear():
    """Clear the context, used in test."""
    _measure_to_view_map_slot.clear()



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from google.rpc import code_pb2

from opencensus.trace import execution_context
from opencensus.trace.status import Status

# By default the excludelist urls are not tracing, currently just include the
# health check url. The paths are literal string matched instead of regular
# expressions. Do not include the '/' at the beginning of the path.
DEFAULT_EXCLUDELIST_PATHS = [
    '_ah/health',
]

# Pattern for matching the 'https://', 'http://', 'ftp://' part.
URL_PATTERN = '^(https?|ftp):\\/\\/'


def get_func_name(func):
    """Return a name which includes the module name and function name."""
    func_name = getattr(func, '__name__', func.__class__.__name__)
    module_name = func.__module__

    if module_name is not None:
        module_name = func.__module__
        return '{}.{}'.format(module_name, func_name)

    return func_name


def disable_tracing_url(url, excludelist_paths=None):
    """Disable tracing on the provided excludelist paths, by default not tracing
    the health check request.

    If the url path starts with the excludelisted path, return True.

    :type excludelist_paths: list
    :param excludelist_paths: Paths that not tracing.

    :rtype: bool
    :returns: True if not tracing, False if tracing.
    """
    if excludelist_paths is None:
        excludelist_paths = DEFAULT_EXCLUDELIST_PATHS

    # Remove the 'https?|ftp://' if exists
    url = re.sub(URL_PATTERN, '', url)

    # Split the url by the first '/' and get the path part
    url_path = url.split('/', 1)[1]

    for path in excludelist_paths:
        if url_path.startswith(path):
            return True

    return False


def disable_tracing_hostname(url, excludelist_hostnames=None):
    """Disable tracing for the provided excludelist URLs, by default not tracing
    the exporter url.

    If the url path starts with the excludelisted path, return True.

    :type excludelist_hostnames: list
    :param excludelist_hostnames: URL that not tracing.

    :rtype: bool
    :returns: True if not tracing, False if tracing.
    """
    if excludelist_hostnames is None:
        # Exporter host_name are not traced by default
        _tracer = execution_context.get_opencensus_tracer()
        try:
            excludelist_hostnames = [
                '{}:{}'.format(
                    _tracer.exporter.host_name,
                    _tracer.exporter.port
                )
            ]
        except(AttributeError):
            excludelist_hostnames = []

    return url in excludelist_hostnames


def status_from_http_code(http_code):
    """Returns equivalent status from http status code
    based on OpenCensus specs.

    :type http_code: int
    :param http_code: HTTP request status code.

    :rtype: int
    :returns: A instance of :class: `~opencensus.trace.status.Status`.
    """
    if http_code <= 199:
        return Status(code_pb2.UNKNOWN)

    if http_code <= 399:
        return Status(code_pb2.OK)

    grpc_code = {
        400: code_pb2.INVALID_ARGUMENT,
        401: code_pb2.UNAUTHENTICATED,
        403: code_pb2.PERMISSION_DENIED,
        404: code_pb2.NOT_FOUND,
        429: code_pb2.RESOURCE_EXHAUSTED,
        501: code_pb2.UNIMPLEMENTED,
        503: code_pb2.UNAVAILABLE,
        504: code_pb2.DEADLINE_EXCEEDED,
    }.get(http_code, code_pb2.UNKNOWN)

    return Status(grpc_code)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common.http_handler import get_request

_GCP_METADATA_URI = 'http://metadata.google.internal/computeMetadata/v1/'
_GCP_METADATA_URI_HEADER = {'Metadata-Flavor': 'Google'}

# ID of the GCP project associated with this resource, such as "my-project"
PROJECT_ID_KEY = 'project_id'

# Numeric VM instance identifier assigned by GCE
INSTANCE_ID_KEY = 'instance_id'

# The GCE zone in which the VM is running
ZONE_KEY = 'zone'

# GKE cluster name
CLUSTER_NAME_KEY = 'instance/attributes/cluster-name'

# GCE common attributes
# See: https://cloud.google.com/appengine/docs/flexible/python/runtime#environment_variables  # noqa
_GCE_ATTRIBUTES = {
    PROJECT_ID_KEY: 'project/project-id',
    INSTANCE_ID_KEY: 'instance/id',
    ZONE_KEY: 'instance/zone'
}

_ATTRIBUTE_URI_TRANSFORMATIONS = {
    _GCE_ATTRIBUTES[ZONE_KEY]:
        lambda v: v[v.rfind('/') + 1:] if '/' in v else v
}

_GCP_METADATA_MAP = {}


class GcpMetadataConfig(object):
    """GcpMetadata represents metadata retrieved from GCP (GKE and GCE)
    environment. Some attributes are retrieved from the system environment.
    see : <a href="https://cloud.google.com/compute/docs/
    storing-retrieving-metadata"> https://cloud.google.com/compute/docs/storing
    -retrieving-metadata</a>
    """
    inited = False
    is_running = False

    @classmethod
    def _initialize_metadata_service(cls):
        """Initialize metadata service once and load gcp metadata into map
        This method should only be called once.
        """
        if cls.inited:
            return

        instance_id = cls.get_attribute('instance/id')

        if instance_id is not None:
            cls.is_running = True

            _GCP_METADATA_MAP['instance_id'] = instance_id

            # fetch attributes from metadata request
            for attribute_key, attribute_uri in _GCE_ATTRIBUTES.items():
                if attribute_key not in _GCP_METADATA_MAP:
                    attribute_value = cls.get_attribute(attribute_uri)
                    if attribute_value is not None:  # pragma: NO COVER
                        _GCP_METADATA_MAP[attribute_key] = attribute_value

        cls.inited = True

    @classmethod
    def is_running_on_gcp(cls):
        cls._initialize_metadata_service()
        return cls.is_running

    def get_gce_metadata(self):
        """for GCP GCE instance"""
        if self.is_running_on_gcp():
            return _GCP_METADATA_MAP

        return dict()

    @staticmethod
    def get_attribute(attribute_uri):
        """
        Fetch the requested instance metadata entry.
        :param attribute_uri: attribute_uri: attribute name relative to the
        computeMetadata/v1 prefix
        :return:  The value read from the metadata service or None
        """
        attribute_value = get_request(_GCP_METADATA_URI + attribute_uri,
                                      _GCP_METADATA_URI_HEADER)

        if attribute_value is not None and isinstance(attribute_value, bytes):
            # At least in python3, bytes are are returned from
            # urllib (although the response is text), convert
            # to a normal string:
            attribute_value = attribute_value.decode('utf-8')

        transformation = _ATTRIBUTE_URI_TRANSFORMATIONS.get(attribute_uri)
        if transformation is not None:
            attribute_value = transformation(attribute_value)

        return attribute_value



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing base class for Span."""


class BaseSpan(object):
    """Base class for Opencensus spans.
    Subclasses of :class:`BaseSpan` must implement the below methods.
    """

    @staticmethod
    def on_create(callback):
        raise NotImplementedError

    @property
    def children(self):
        """The child spans of the current span."""
        raise NotImplementedError

    def span(self, name='child_span'):
        """Create a child span for the current span and append it to the child
        spans list.

        :type name: str
        :param name: (Optional) The name of the child span.

        :rtype: :class: `~opencensus.trace.span.Span`
        :returns: A child Span to be added to the current span.
        """
        raise NotImplementedError

    def add_attribute(self, attribute_key, attribute_value):
        """Add attribute to span.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        raise NotImplementedError

    def add_annotation(self, description, **attrs):
        """Add an annotation to span.

        :type description: str
        :param description: A user-supplied message describing the event.
                        The maximum length for the description is 256 bytes.

        :type attrs: kwargs
        :param attrs: keyworded arguments e.g. failed=True, name='Caching'
        """
        raise NotImplementedError

    def add_message_event(self, message_event):
        """Add a message event to this span.

        :type message_event: :class:`opencensus.trace.time_event.MessageEvent`
        :param message_event: The message event to attach to this span.
        """
        raise NotImplementedError

    def add_link(self, link):
        """Add a Link.

        :type link: :class: `~opencensus.trace.link.Link`
        :param link: A Link object.
        """
        raise NotImplementedError

    def set_status(self, status):
        """Sets span status.

        :type code: :class: `~opencensus.trace.status.Status`
        :param code: A Status object.
        """
        raise NotImplementedError

    def start(self):
        """Set the start time for a span."""
        raise NotImplementedError

    def finish(self):
        """Set the end time for a span."""
        raise NotImplementedError

    def __iter__(self):
        """Iterate through the span tree."""
        raise NotImplementedError

    def __enter__(self):
        """Start a span."""
        raise NotImplementedError

    def __exit__(self, exception_type, exception_value, traceback):
        """Finish a span."""
        raise NotImplementedError



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common import utils


class ViewData(object):
    """View Data is the aggregated data for a particular view

    :type view:
    :param view: The view associated with this view data

    :type start_time: datetime
    :param start_time: the start time for this view data

    :type end_time: datetime
    :param end_time: the end time for this view data

    """
    def __init__(self,
                 view,
                 start_time,
                 end_time):
        self._view = view
        self._start_time = start_time
        self._end_time = end_time
        self._tag_value_aggregation_data_map = {}

    @property
    def view(self):
        """the current view in the view data"""
        return self._view

    # TODO: `start_time` and `end_time` are sometimes a `datetime` object but
    # should always be a `string`.
    @property
    def start_time(self):
        """the current start time in the view data"""
        return self._start_time

    @property
    def end_time(self):
        """the current end time in the view data"""
        return self._end_time

    @property
    def tag_value_aggregation_data_map(self):
        """the current tag value aggregation map in the view data"""
        return self._tag_value_aggregation_data_map

    def start(self):
        """sets the start time for the view data"""
        self._start_time = utils.to_iso_str()

    def end(self):
        """sets the end time for the view data"""
        self._end_time = utils.to_iso_str()

    def get_tag_values(self, tags, columns):
        """function to get the tag values from tags and columns"""
        tag_values = []
        i = 0
        while i < len(columns):
            tag_key = columns[i]
            if tag_key in tags:
                tag_values.append(tags.get(tag_key))
            else:
                tag_values.append(None)
            i += 1
        return tag_values

    def record(self, context, value, timestamp, attachments=None):
        """records the view data against context"""
        if context is None:
            tags = dict()
        else:
            tags = context.map
        tag_values = self.get_tag_values(tags=tags,
                                         columns=self.view.columns)
        tuple_vals = tuple(tag_values)
        if tuple_vals not in self.tag_value_aggregation_data_map:
            self.tag_value_aggregation_data_map[tuple_vals] = \
                self.view.new_aggregation_data()
        self.tag_value_aggregation_data_map.get(tuple_vals).\
            add_sample(value, timestamp, attachments)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.trace.span_context import SpanContext
from opencensus.trace.trace_options import TraceOptions

_OPENCENSUS_TRACE_PREFIX = 'opencensus-trace'
_TRACE_ID_KEY = '{}-traceid'.format(_OPENCENSUS_TRACE_PREFIX)
_SPAN_ID_KEY = '{}-spanid'.format(_OPENCENSUS_TRACE_PREFIX)
_TRACE_OPTIONS_KEY = '{}-traceoptions'.format(_OPENCENSUS_TRACE_PREFIX)

DEFAULT_TRACE_OPTIONS = '1'


class TextFormatPropagator(object):
    """This class provides the basic utilities for extracting the trace
    information from a carrier which is a dict to form a SpanContext. And
    generating a dict using the provided SpanContext.
    """
    def from_carrier(self, carrier):
        """Generate a SpanContext object using the information in the carrier.

        :type carrier: dict
        :param carrier: The carrier which has the trace_id, span_id, options
                        information for creating a SpanContext.

        :rtype: :class:`~opencensus.trace.span_context.SpanContext`
        :returns: SpanContext generated from the carrier.
        """
        trace_id = None
        span_id = None
        trace_options = None

        for key in carrier:
            key = key.lower()
            if key == _TRACE_ID_KEY:
                trace_id = carrier[key]
            if key == _SPAN_ID_KEY:
                span_id = carrier[key]
            if key == _TRACE_OPTIONS_KEY:
                trace_options = bool(carrier[key])

        if trace_options is None:
            trace_options = DEFAULT_TRACE_OPTIONS

        return SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            trace_options=TraceOptions(trace_options),
            from_header=True)

    def to_carrier(self, span_context, carrier):
        """Inject the SpanContext fields to carrier dict.

        :type span_context:
            :class:`~opencensus.trace.span_context.SpanContext`
        :param span_context: SpanContext object.

        :type carrier: dict
        :param carrier: The carrier which holds the trace_id, span_id, options
                        information from a SpanContext.

        :rtype: dict
        :returns: The carrier which holds the span context information.
        """
        carrier[_TRACE_ID_KEY] = str(span_context.trace_id)

        if span_context.span_id is not None:
            carrier[_SPAN_ID_KEY] = str(span_context.span_id)

        carrier[_TRACE_OPTIONS_KEY] = str(
            span_context.trace_options.trace_options_byte)

        return carrier



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import namedtuple
from copy import copy

from opencensus.trace import execution_context

_meta_logger = logging.getLogger(__name__)

TRACE_ID_KEY = 'traceId'
SPAN_ID_KEY = 'spanId'
SAMPLING_DECISION_KEY = 'traceSampled'

LogAttrs = namedtuple('LogAttrs', ['trace_id', 'span_id', 'sampling_decision'])
ATTR_DEFAULTS = LogAttrs("00000000000000000000000000000000",
                         "0000000000000000", False)


def get_log_attrs():
    """Get logging attributes from the opencensus context.

    :rtype: :class:`LogAttrs`
    :return: The current span's trace ID, span ID, and sampling decision.
    """
    try:
        tracer = execution_context.get_opencensus_tracer()
        if tracer is None:
            raise RuntimeError
    except Exception:  # noqa
        _meta_logger.error("Failed to get opencensus tracer")
        return ATTR_DEFAULTS

    try:
        trace_id = tracer.span_context.trace_id
        if trace_id is None:
            trace_id = ATTR_DEFAULTS.trace_id
    except Exception:  # noqa
        _meta_logger.error("Failed to get opencensus trace ID")
        trace_id = ATTR_DEFAULTS.trace_id

    try:
        span_id = tracer.span_context.span_id
        if span_id is None:
            span_id = ATTR_DEFAULTS.span_id
    except Exception:  # noqa
        _meta_logger.error("Failed to get opencensus span ID")
        span_id = ATTR_DEFAULTS.span_id

    try:
        sampling_decision = tracer.span_context.trace_options.get_enabled()
        if sampling_decision is None:
            sampling_decision = ATTR_DEFAULTS.sampling_decision
    except AttributeError:
        sampling_decision = ATTR_DEFAULTS.sampling_decision
    except Exception:  # noqa
        _meta_logger.error("Failed to get opencensus sampling decision")
        sampling_decision = ATTR_DEFAULTS.sampling_decision

    return LogAttrs(trace_id, span_id, sampling_decision)


def _set_extra_attrs(extra):
    trace_id, span_id, sampling_decision = get_log_attrs()
    extra.setdefault(TRACE_ID_KEY, trace_id)
    extra.setdefault(SPAN_ID_KEY, span_id)
    extra.setdefault(SAMPLING_DECISION_KEY, sampling_decision)


# See
# https://docs.python.org/3.7/library/logging.html#loggeradapter-objects,
# https://docs.python.org/3.7/howto/logging-cookbook.html#context-info
class TraceLoggingAdapter(logging.LoggerAdapter):
    """Adapter to add opencensus context attrs to records."""
    def process(self, msg, kwargs):
        kwargs = copy(kwargs)
        if self.extra:
            extra = copy(self.extra)
        else:
            extra = {}
        extra.update(kwargs.get('extra', {}))
        _set_extra_attrs(extra)
        kwargs['extra'] = extra

        return (msg, kwargs)


# This is the idiomatic way to stack logger customizations, see
# https://docs.python.org/3.7/library/logging.html#logging.getLoggerClass
class TraceLogger(logging.getLoggerClass()):
    """Logger class that adds opencensus context attrs to records."""
    def makeRecord(self, *args, **kwargs):
        try:
            extra = args[8]
            if extra is None:
                extra = {}
                args = tuple(list(args[:8]) + [extra] + list(args[9:]))
        except IndexError:  # pragma: NO COVER
            extra = kwargs.setdefault('extra', {})
            if extra is None:
                kwargs['extra'] = extra
        _set_extra_attrs(extra)
        return super(TraceLogger, self).makeRecord(*args, **kwargs)



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading


class MetricProducer(object):
    """Produces a set of metrics for export."""

    def get_metrics(self):
        """Get a set of metrics to be exported.

        :rtype: set(:class: `opencensus.metrics.export.metric.Metric`)
        :return: A set of metrics to be exported.
        """
        raise NotImplementedError  # pragma: NO COVER


class MetricProducerManager(object):
    """Container class for MetricProducers to be used by exporters.

    :type metric_producers: iterable(class: 'MetricProducer')
    :param metric_producers: Optional initial metric producers.
    """

    def __init__(self, metric_producers=None):
        if metric_producers is None:
            self.metric_producers = set()
        else:
            self.metric_producers = set(metric_producers)
        self.mp_lock = threading.Lock()

    def add(self, metric_producer):
        """Add a metric producer.

        :type metric_producer: :class: 'MetricProducer'
        :param metric_producer: The metric producer to add.
        """
        if metric_producer is None:
            raise ValueError
        with self.mp_lock:
            self.metric_producers.add(metric_producer)

    def remove(self, metric_producer):
        """Remove a metric producer.

        :type metric_producer: :class: 'MetricProducer'
        :param metric_producer: The metric producer to remove.
        """
        if metric_producer is None:
            raise ValueError
        try:
            with self.mp_lock:
                self.metric_producers.remove(metric_producer)
        except KeyError:
            pass

    def get_all(self):
        """Get the set of all metric producers.

        Get a copy of `metric_producers`. Prefer this method to using the
        attribute directly to avoid other threads adding/removing producers
        while you're reading it.

        :rtype: set(:class: `MetricProducer`)
        :return: A set of all metric producers at the time of the call.
        """
        with self.mp_lock:
            mps_copy = set(self.metric_producers)
        return mps_copy



# Copyright 2019, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import queue

import logging
import threading
import time

logger = logging.getLogger(__name__)


class PeriodicTask(threading.Thread):
    """Thread that periodically calls a given function.

    :type interval: int or float
    :param interval: Seconds between calls to the function.

    :type function: function
    :param function: The function to call.

    :type args: list
    :param args: The args passed in while calling `function`.

    :type kwargs: dict
    :param kwargs: The kwargs passed in while calling `function`.

    :type name: str
    :param name: The source of the worker. Used for naming.
    """

    def __init__(self, interval, function, args=None, kwargs=None, name=None):
        super(PeriodicTask, self).__init__(name=name)
        self.interval = interval
        self.function = function
        self.args = args or []
        self.kwargs = kwargs or {}
        self.finished = threading.Event()

    def run(self):
        wait_time = self.interval
        while not self.finished.wait(wait_time):
            start_time = time.time()
            self.function(*self.args, **self.kwargs)
            elapsed_time = time.time() - start_time
            wait_time = max(self.interval - elapsed_time, 0)

    def cancel(self):
        self.finished.set()


class QueueEvent(object):
    def __init__(self, name):
        self.name = name
        self.event = threading.Event()

    def __repr__(self):
        return ('{}({})'.format(type(self).__name__, self.name))

    def set(self):
        return self.event.set()

    def wait(self, timeout=None):
        return self.event.wait(timeout)


class QueueExitEvent(QueueEvent):
    pass


class Queue(object):
    def __init__(self, capacity):
        self.EXIT_EVENT = QueueExitEvent('EXIT')
        self._queue = queue.Queue(maxsize=capacity)

    def _gets(self, count, timeout):
        start_time = time.time()
        elapsed_time = 0
        cnt = 0
        while cnt < count:
            try:
                item = self._queue.get(block=False)
                yield item
                if isinstance(item, QueueEvent):
                    return
            except queue.Empty:
                break
            cnt += 1
        while cnt < count:
            wait_time = max(timeout - elapsed_time, 0)
            try:
                item = self._queue.get(block=True, timeout=wait_time)
                yield item
                if isinstance(item, QueueEvent):
                    return
            except queue.Empty:
                break
            cnt += 1
            elapsed_time = time.time() - start_time

    def gets(self, count, timeout):
        return tuple(self._gets(count, timeout))

    def is_empty(self):
        return not self._queue.qsize()

    def flush(self, timeout=None):
        if self._queue.qsize() == 0:
            return 0
        start_time = time.time()
        wait_time = timeout
        event = QueueEvent('SYNC(timeout={})'.format(wait_time))
        try:
            self._queue.put(event, block=True, timeout=wait_time)
        except queue.Full:
            return
        elapsed_time = time.time() - start_time
        wait_time = timeout and max(timeout - elapsed_time, 0)
        if event.wait(wait_time):
            return time.time() - start_time  # time taken to flush

    def put(self, item, block=True, timeout=None):
        try:
            self._queue.put(item, block, timeout)
        except queue.Full:
            logger.warning('Queue is full. Dropping telemetry.')

    def puts(self, items, block=True, timeout=None):
        if block and timeout is not None:
            start_time = time.time()
            elapsed_time = 0
            for item in items:
                wait_time = max(timeout - elapsed_time, 0)
                self.put(item, block=True, timeout=wait_time)
                elapsed_time = time.time() - start_time
        else:
            for item in items:
                self.put(item, block, timeout)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from opencensus.tags.tag_key import TagKey
from opencensus.tags.tag_value import TagValue


class TagMap(object):
    """ A tag map is a map of tags from key to value

    :type tags: list(:class: '~opencensus.tags.tag.Tag')
    :param tags: a list of tags

    """

    def __init__(self, tags=None):
        self.map = OrderedDict(tags if tags else [])

    def __iter__(self):
        return self.map.items().__iter__()

    def insert(self, key, value):
        """Inserts a key and value in the map if the map does not already
        contain the key.

        :type key: :class: '~opencensus.tags.tag_key.TagKey'
        :param key: a tag key to insert into the map

        :type value: :class: '~opencensus.tags.tag_value.TagValue'
        :param value: a tag value that is associated with the tag key and
        the value to insert into the tag map

        """
        if key in self.map:
            return

        try:
            tag_key = TagKey(key)
            tag_val = TagValue(value)
            self.map[tag_key] = tag_val
        except ValueError:
            raise

    def delete(self, key):
        """Deletes a tag from the map if the key is in the map

        :type key: :class: '~opencensus.tags.tag_key.TagKey'
        :param key: A string representing a possible tag key

        :returns: the value of the key in the dictionary if it is in there,
                  or None if it is not.
        """
        self.map.pop(key, None)

    def update(self, key, value):
        """Updates the map by updating the value of a key

        :type key: :class: '~opencensus.tags.tag_key.TagKey'
        :param key: A tag key to be updated

        :type value: :class: '~opencensus.tags.tag_value.TagValue'
        :param value: The value to update the key to in the map

        """
        if key in self.map:
            self.map[key] = value

    def tag_key_exists(self, key):
        """Checking if the tag key exists in the map

        :type key: '~opencensus.tags.tag_key.TagKey'
        :param key: A string to check to see if that is a key in the map

        :returns: True if the key is in map, False is it is not

        """
        return key in self.map

    def get_value(self, key):
        """ Gets the value of the key passed in if the key exists in the map

        :type key: str
        :param key: A string representing a key to get the value of in the map

        :returns: A KeyError if the value is None, else returns the value

        """
        try:
            return self.map[key]
        except KeyError:
            raise KeyError('key is not in map')



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

from opencensus.tags.tag_key import TagKey
from opencensus.tags.tag_value import TagValue

Tag_ = namedtuple('Tag', ['key', 'value'])


class Tag(Tag_):
    """A tag, in the format [KEY]:[VALUE].

    :type key: str
    :param key: The name of the tag

    :type value: str
    :param value: The value of the tag

    """
    def __new__(cls, key, value):
        return super(Tag, cls).__new__(
            cls,
            key=TagKey(key),
            value=TagValue(value),
        )



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.trace import base_span
from opencensus.trace.span_context import generate_span_id
from opencensus.trace.tracers import base


class BlankSpan(base_span.BaseSpan):
    """A BlankSpan is an individual timed event which forms a node of the trace
    tree. All operations are no-op.

    :type name: str
    :param name: The name of the span.

    :type parent_span: :class:`~opencensus.trace.blank_span.BlankSpan`
    :param parent_span: (Optional) Parent span.

    :type status: :class: `~opencensus.trace.status.Status`
    :param status: (Optional) An optional final status for this span.

    :type context_tracer: :class:`~opencensus.trace.tracers.noop_tracer.
                                 NoopTracer`
    :param context_tracer: The tracer that holds a stack of spans. If this is
                           not None, then when exiting a span, use the end_span
                           method in the tracer class to finish a span. If no
                           tracer is passed in, then just finish the span using
                           the finish method in the Span class.
    """

    def __init__(
            self,
            name=None,
            parent_span=None,
            attributes=None,
            start_time=None,
            end_time=None,
            span_id=None,
            stack_trace=None,
            annotations=None,
            message_events=None,
            links=None,
            status=None,
            same_process_as_parent_span=None,
            context_tracer=None,
            span_kind=None):
        self.name = name
        self.parent_span = parent_span
        self.start_time = start_time
        self.end_time = end_time

        self.span_id = generate_span_id()
        self.parent_span = base.NullContextManager()

        self.attributes = {}
        self.stack_trace = stack_trace
        self.annotations = annotations
        self.message_events = message_events
        self.links = []
        self.status = status
        self.same_process_as_parent_span = same_process_as_parent_span
        self._child_spans = []
        self.context_tracer = context_tracer
        self.span_kind = span_kind

    @staticmethod
    def on_create(callback):
        pass

    @property
    def children(self):
        """The child spans of the current BlankSpan."""
        return list()

    def span(self, name='child_span'):
        """Create a child span for the current span and append it to the child
        spans list.

        :type name: str
        :param name: (Optional) The name of the child span.

        :rtype: :class: `~opencensus.trace.blankspan.BlankSpan`
        :returns: A child Span to be added to the current span.
        """
        child_span = BlankSpan(name, parent_span=self)
        self._child_spans.append(child_span)
        return child_span

    def add_attribute(self, attribute_key, attribute_value):
        """No-op implementation of this method.

        :type attribute_key: str
        :param attribute_key: Attribute key.

        :type attribute_value:str
        :param attribute_value: Attribute value.
        """
        pass

    def add_annotation(self, description, **attrs):
        """No-op implementation of this method.

        :type description: str
        :param description: A user-supplied message describing the event.
                        The maximum length for the description is 256 bytes.

        :type attrs: kwargs
        :param attrs: keyworded arguments e.g. failed=True, name='Caching'
        """
        pass

    def add_message_event(self, message_event):
        """No-op implementation of this method.

        :type message_event: :class:`opencensus.trace.time_event.MessageEvent`
        :param message_event: The message event to attach to this span.
        """
        pass

    def add_link(self, link):
        """No-op implementation of this method.

        :type link: :class: `~opencensus.trace.link.Link`
        :param link: A Link object.
        """
        pass

    def set_status(self, status):
        """No-op implementation of this method.

        :type code: :class: `~opencensus.trace.status.Status`
        :param code: A Status object.
        """
        pass

    def start(self):
        """No-op implementation of this method."""
        pass

    def finish(self):
        """No-op implementation of this method."""
        pass

    def __iter__(self):
        """Iterate through the span tree."""
        yield self

    def __enter__(self):
        """Start a span."""
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Finish a span."""
        pass



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.stats import execution_context
from opencensus.stats.measure_to_view_map import MeasureToViewMap
from opencensus.stats.measurement_map import MeasurementMap


class StatsRecorder(object):
    """Stats Recorder provides methods to record stats against tags

    """
    def __init__(self):
        if execution_context.get_measure_to_view_map() == {}:
            execution_context.set_measure_to_view_map(MeasureToViewMap())

        self.measure_to_view_map = execution_context.get_measure_to_view_map()

    def new_measurement_map(self):
        """Creates a new MeasurementMap in order to record stats
        :returns a MeasurementMap for recording multiple measurements
        """
        return MeasurementMap(self.measure_to_view_map)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SpanContext encapsulates the current context within the request's trace."""

import six

import logging
import random
import re

from opencensus.trace import trace_options as trace_options_module

_INVALID_TRACE_ID = '0' * 32
INVALID_SPAN_ID = '0' * 16

TRACE_ID_PATTERN = re.compile('[0-9a-f]{32}?')
SPAN_ID_PATTERN = re.compile('[0-9a-f]{16}?')

# Default options, don't force sampling
DEFAULT_OPTIONS = '0'


class SpanContext(object):
    """SpanContext includes 3 fields: traceId, spanId, and an trace_options flag
    which indicates whether or not the request is being traced. It contains the
    current context to be propagated to the child spans.

    :type trace_id: str
    :param trace_id: (Optional) Trace_id is a 32 digits uuid for the trace.
                     If not given, will generate one automatically.

    :type span_id: str
    :param span_id: (Optional) Identifier for the span, unique within a trace.

    :type trace_options: :class: `~opencensus.trace.trace_options.TraceOptions`
    :param trace_options: (Optional) TraceOptions indicates 8 trace options.

    :type from_header: bool
    :param from_header: (Optional) Indicates whether the trace context is
                        generated from request header.
    """
    def __init__(
            self,
            trace_id=None,
            span_id=None,
            trace_options=None,
            tracestate=None,
            from_header=False):
        if trace_id is None:
            trace_id = generate_trace_id()

        if trace_options is None:
            trace_options = trace_options_module.TraceOptions(DEFAULT_OPTIONS)

        self.from_header = from_header
        self.trace_id = self._check_trace_id(trace_id)
        self.span_id = self._check_span_id(span_id)
        self.trace_options = trace_options
        self.tracestate = tracestate

    def __repr__(self):
        """Returns a string form of the SpanContext.

        :rtype: str
        :returns: String form of the SpanContext.
        """
        fmt = '{}(trace_id={}, span_id={}, trace_options={}, tracestate={})'
        return fmt.format(
            type(self).__name__,
            self.trace_id,
            self.span_id,
            self.trace_options,
            self.tracestate,
        )

    def _check_span_id(self, span_id):
        """Check the format of the span_id to ensure it is 16-character hex
        value representing a 64-bit number. If span_id is invalid, logs a
        warning message and returns None

        :type span_id: str
        :param span_id: Identifier for the span, unique within a span.

        :rtype: str
        :returns: Span_id for the current span.
        """
        if span_id is None:
            return None
        assert isinstance(span_id, six.string_types)

        if span_id is INVALID_SPAN_ID:
            logging.warning(
                'Span_id %s is invalid (cannot be all zero)', span_id)
            self.from_header = False
            return None

        match = SPAN_ID_PATTERN.match(span_id)

        if match:
            return span_id
        else:
            logging.warning(
                'Span_id %s does not the match the '
                'required format', span_id)
            self.from_header = False
            return None

    def _check_trace_id(self, trace_id):
        """Check the format of the trace_id to ensure it is 32-character hex
        value representing a 128-bit number. If trace_id is invalid, returns a
        randomly generated trace id

        :type trace_id: str
        :param trace_id:

        :rtype: str
        :returns: Trace_id for the current context.
        """
        assert isinstance(trace_id, six.string_types)

        if trace_id is _INVALID_TRACE_ID:
            logging.warning(
                'Trace_id %s is invalid (cannot be all zero), '
                'generating a new one.', trace_id)
            self.from_header = False
            return generate_trace_id()

        match = TRACE_ID_PATTERN.match(trace_id)

        if match:
            return trace_id
        else:
            logging.warning(
                'Trace_id %s does not the match the required format,'
                'generating a new one instead.', trace_id)
            self.from_header = False
            return generate_trace_id()


def generate_span_id():
    """Return the random generated span ID for a span. Must be a 16 character
    hexadecimal encoded string

    :rtype: str
    :returns: 16 digit randomly generated hex trace id.
    """
    return '{:016x}'.format(random.getrandbits(64))


def generate_trace_id():
    """Generate a random 32 char hex trace_id.

    :rtype: str
    :returns: 32 digit randomly generated hex trace id.
    """
    return '{:032x}'.format(random.getrandbits(128))



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseMeasure(object):
    """ A measure is the type of metric that is being recorded with
    a name, description, and unit

    :type name: str
    :param name: string representing the name of the measure

    :type description: str
    :param description: a string representing the description of the measure

    :type unit: str
    :param unit: the units in which the measure values are measured

    """
    def __init__(self, name, description, unit=None):
        self._name = name
        self._description = description
        self._unit = unit

    @property
    def name(self):
        """The name of the current measure"""
        return self._name

    @property
    def description(self):
        """The description of the current measure"""
        return self._description

    @property
    def unit(self):
        """The unit of the current measure"""
        return self._unit


class MeasureInt(BaseMeasure):
    """Creates an Integer Measure"""
    def __init__(self, name, description, unit=None):
        super(MeasureInt, self).__init__(name, description, unit)


class MeasureFloat(BaseMeasure):
    """Creates a Float Measure"""
    def __init__(self, name, description, unit=None):
        super(MeasureFloat, self).__init__(name, description, unit)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities to convert stats data models to metrics data models.
"""

from opencensus.metrics import label_value
from opencensus.metrics.export import metric, metric_descriptor, time_series


def is_gauge(md_type):
    """Whether a given MetricDescriptorType value is a gauge.

    :type md_type: int
    :param md_type: A MetricDescriptorType enum value.
    """
    if md_type not in metric_descriptor.MetricDescriptorType:
        raise ValueError  # pragma: NO COVER

    return md_type in {
        metric_descriptor.MetricDescriptorType.GAUGE_INT64,
        metric_descriptor.MetricDescriptorType.GAUGE_DOUBLE,
        metric_descriptor.MetricDescriptorType.GAUGE_DISTRIBUTION
    }


def get_label_values(tag_values):
    """Convert an iterable of TagValues into a list of LabelValues.

    :type tag_values: list(:class: `opencensus.tags.tag_value.TagValue`)
    :param tag_values: An iterable of TagValues to convert.

    :rtype: list(:class: `opencensus.metrics.label_value.LabelValue`)
    :return: A list of LabelValues, converted from TagValues.
    """
    return [label_value.LabelValue(tv) for tv in tag_values]


def view_data_to_metric(view_data, timestamp):
    """Convert a ViewData to a Metric at time `timestamp`.

    :type view_data: :class: `opencensus.stats.view_data.ViewData`
    :param view_data: The ViewData to convert.

    :type timestamp: :class: `datetime.datetime`
    :param timestamp: The time to set on the metric's point's aggregation,
    usually the current time.

    :rtype: :class: `opencensus.metrics.export.metric.Metric`
    :return: A converted Metric.
    """
    if not view_data.tag_value_aggregation_data_map:
        return None

    md = view_data.view.get_metric_descriptor()

    # TODO: implement gauges
    if is_gauge(md.type):
        ts_start = None  # pragma: NO COVER
    else:
        ts_start = view_data.start_time

    ts_list = []
    for tag_vals, agg_data in view_data.tag_value_aggregation_data_map.items():
        label_values = get_label_values(tag_vals)
        point = agg_data.to_point(timestamp)
        ts_list.append(time_series.TimeSeries(label_values, [point], ts_start))
    return metric.Metric(md, ts_list)



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import random
import traceback

from opencensus.common.utils import get_truncatable_str

MAX_FRAMES = 128

BUILD_ID = os.environ.get('BUILD_ID', 'unknown')
SOURCE_VERSION = os.environ.get('SOURCE_VERSION', 'unknown')


class StackFrame(object):
    """Represents a single stack frame in a stack trace.

    :type func_name: str
    :param func_name: The fully-qualified name that uniquely identifies the
                      function or method that is active in this frame (up to
                      1024 bytes).

    :type original_func_name: str
    :param original_func_name: An un-mangled function name, if functionName is
                               mangled. The name can be fully-qualified
                               (up to 1024 bytes).

    :type file_name: str
    :param file_name: The name of the source file where the function call
                      appears (up to 256 bytes).

    :type line_num: int
    :param line_num: The line number in fileName where the function call
                     appears.

    :type col_num: int
    :param col_num: The column number where the function call appears, if
                    available. This is important in JavaScript because of its
                    anonymous functions.

    :type load_module: str
    :param load_module: For example: main binary, kernel modules, and dynamic
                        libraries such as libc.so, sharedlib.so
                        (up to 256 bytes).

    :type build_id: str
    :param build_id: A unique identifier for the module, usually a hash of its
                    contents (up to 128 bytes).


    :type source_version: str
    :param source_version: The version of the deployed source code
                           (up to 128 bytes).
    """
    def __init__(self,
                 func_name,
                 original_func_name,
                 file_name,
                 line_num,
                 col_num,
                 load_module,
                 build_id,
                 source_version):
        self.func_name = func_name
        self.original_func_name = original_func_name
        self.file_name = file_name
        self.line_num = line_num
        self.col_num = col_num
        self.load_module = load_module
        self.build_id = build_id
        self.source_version = source_version

    def format_stack_frame_json(self):
        """Convert StackFrame object to json format."""
        stack_frame_json = {}
        stack_frame_json['function_name'] = get_truncatable_str(
            self.func_name)
        stack_frame_json['original_function_name'] = get_truncatable_str(
            self.original_func_name)
        stack_frame_json['file_name'] = get_truncatable_str(self.file_name)
        stack_frame_json['line_number'] = self.line_num
        stack_frame_json['column_number'] = self.col_num
        stack_frame_json['load_module'] = {
            'module': get_truncatable_str(self.load_module),
            'build_id': get_truncatable_str(self.build_id),
        }
        stack_frame_json['source_version'] = get_truncatable_str(
            self.source_version)

        return stack_frame_json


class StackTrace(object):
    """A call stack appearing in a trace.

    :type stack_frames: list
    :param stack_frames: Stack frames in this stack trace. A maximum of 128
                         frames are allowed.

    :type stack_trace_hash_id: str
    :param stack_trace_hash_id: The hash ID is used to conserve network
                                bandwidth for duplicate stack traces within a
                                single trace.
    """
    def __init__(self, stack_frames=None, stack_trace_hash_id=None):
        if stack_frames is None:
            stack_frames = []
        if len(stack_frames) > MAX_FRAMES:
            self.dropped_frames_count = len(stack_frames) - MAX_FRAMES
            stack_frames = stack_frames[-MAX_FRAMES:]
        else:
            self.dropped_frames_count = 0

        if stack_trace_hash_id is None:
            stack_trace_hash_id = generate_hash_id()

        self.stack_frames = stack_frames
        self.stack_trace_hash_id = stack_trace_hash_id

    @classmethod
    def from_traceback(cls, tb):
        """Initializes a StackTrace from a python traceback instance"""
        stack_trace = cls(
            stack_trace_hash_id=generate_hash_id_from_traceback(tb)
        )
        # use the add_stack_frame so that json formatting is applied
        for tb_frame_info in traceback.extract_tb(tb):
            filename, line_num, fn_name, _ = tb_frame_info
            stack_trace.add_stack_frame(
                StackFrame(
                    func_name=fn_name,
                    original_func_name=fn_name,
                    file_name=filename,
                    line_num=line_num,
                    col_num=0,  # I don't think this is available in python
                    load_module=filename,
                    build_id=BUILD_ID,
                    source_version=SOURCE_VERSION
                )
            )
        return stack_trace

    def add_stack_frame(self, stack_frame):
        """Add StackFrame to frames list."""
        if len(self.stack_frames) >= MAX_FRAMES:
            self.dropped_frames_count += 1
        else:
            self.stack_frames.append(stack_frame.format_stack_frame_json())

    def format_stack_trace_json(self):
        """Convert a StackTrace object to json format."""
        stack_trace_json = {}

        if self.stack_frames:
            stack_trace_json['stack_frames'] = {
                'frame': self.stack_frames,
                'dropped_frames_count': self.dropped_frames_count
            }

        stack_trace_json['stack_trace_hash_id'] = self.stack_trace_hash_id

        return stack_trace_json


def generate_hash_id():
    """Generate a hash id."""
    return random.getrandbits(64)


def generate_hash_id_from_traceback(tb):
    m = hashlib.md5()  # nosec
    for tb_line in traceback.format_tb(tb):
        m.update(tb_line.encode('utf-8'))
    # truncate the hash for easier compatibility with StackDriver,
    # should still be unique enough to avoid collisions
    return int(m.hexdigest()[:12], 16)



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The classes in this module implement the spec for v1 Metrics as of
opencensus-proto release v0.1.0. See opencensus-proto for details:

https://github.com/census-instrumentation/opencensus-proto/blob/v0.1.0/src/opencensus/proto/metrics/v1/metrics.proto
"""  # noqa

from copy import copy


class ValueDouble(object):
    """A 64-bit double-precision floating-point number.

    :type value: float
    :param value: the value in float.
    """

    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    @property
    def value(self):
        return self._value


class ValueLong(object):
    """A 64-bit integer.

    :type value: long
    :param value: the value in long.
    """

    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    @property
    def value(self):
        return self._value


class ValueSummary(object):
    """Represents a snapshot values calculated over an arbitrary time window.

    :type value: summary
    :param value: the value in summary.
    """

    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    @property
    def value(self):
        return self._value


class Exemplar(object):
    """An example point to annotate a given value in a bucket.

    Exemplars are example points that may be used to annotate aggregated
    Distribution values. They are metadata that gives information about a
    particular value added to a Distribution bucket.

    :type value: double
    :param value: Value of the exemplar point, determines which bucket the
    exemplar belongs to.

    :type timestamp: str
    :param timestamp: The observation (sampling) time of the exemplar value.

    :type attachments: dict(str, str)
    :param attachments: Contextual information about the example value.
    """

    def __init__(self, value, timestamp, attachments):
        self._value = value
        self._timestamp = timestamp
        self._attachments = attachments

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.value,
                ))

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def attachments(self):
        return self._attachments


class Bucket(object):
    """A bucket of a histogram.

    :type count: int
    :param count: The number of values in each bucket of the histogram.

    :type exemplar: Exemplar
    :param exemplar: Optional exemplar for this bucket, omit if the
    distribution does not have a histogram.
    """

    def __init__(self, count, exemplar=None):
        self._count = count
        self._exemplar = exemplar

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.count,
                ))

    @property
    def count(self):
        return self._count

    @property
    def exemplar(self):
        return self._exemplar


class Explicit(object):
    """Set of explicit bucket boundaries.

    Specifies a set of buckets with arbitrary upper-bounds.  This defines
    size(bounds) + 1 (= N) buckets. The boundaries for bucket index i are:

        - [0, bounds[i]) for i == 0
        - [bounds[i-1], bounds[i]) for 0 < i < N-1
        - [bounds[i-1], +infinity) for i == N-1
    """

    def __init__(self, bounds):
        if not bounds:
            raise ValueError("Bounds must not be null or empty")
        if bounds != sorted(set(bounds)):
            raise ValueError("Bounds must be strictly increasing")
        if bounds[0] <= 0:
            raise ValueError("Bounds must be positive")
        self._bounds = bounds

    @property
    def bounds(self):
        return copy(self._bounds)


class BucketOptions(object):
    """Container for bucket options, including explicit boundaries.

    A Distribution may optionally contain a histogram of the values in the
    population. The bucket boundaries for that histogram are described by
    BucketOptions.

    If bucket_options has no type, then there is no histogram associated with
    the Distribution.
    """

    def __init__(self, type_=None):
        self._type = type_

    def __repr__(self):
        return ("{}({})"
                .format(
                    type(self).__name__,
                    self.type_,
                ))

    @property
    def type_(self):
        return self._type


class ValueDistribution(object):
    """Summary statistics for a population of values.

    Distribution contains summary statistics for a population of values. It
    optionally contains a histogram representing the distribution of those
    values across a set of buckets.

    :type count: int
    :param count: The number of values in the population.

    :type sum_: float
    :param sum_: The sum of the values in the population.

    :type sum_of_squared_deviation: float
    :param sum_of_squared_deviation: The sum of squared deviations from the
    mean of the values in the population.

    :type bucket_options: :class: 'BucketOptions'
    :param bucket_options: Bucket boundaries for the histogram of the values in
    the population.

    :type buckets: list(:class: 'Bucket')
    :param buckets: Histogram buckets for the given bucket boundaries.
    """

    def __init__(self,
                 count,
                 sum_,
                 sum_of_squared_deviation,
                 bucket_options,
                 buckets=None):
        if count < 0:
            raise ValueError("count must be non-negative")
        elif count == 0:
            if sum_ != 0:
                raise ValueError("sum_ must be 0 if count is 0")
            if sum_of_squared_deviation != 0:
                raise ValueError("sum_of_squared_deviation must be 0 if count "
                                 "is 0")
        if bucket_options is None:
            raise ValueError("bucket_options must not be null")
        if bucket_options.type_ is None:
            if buckets is not None:
                raise ValueError("buckets must be null if the distribution "
                                 "has no histogram (i.e. bucket_options.type "
                                 "is null)")
        else:
            if len(buckets) != len(bucket_options.type_.bounds) + 1:
                # Note that this includes the implicit 0 and positive-infinity
                # boundaries, so bounds [1, 2] implies three buckets: [[0, 1),
                # [1, 2), [2, inf)].
                raise ValueError("There must be one bucket for each pair of "
                                 "boundaries")
            if count != sum(bucket.count for bucket in buckets):
                raise ValueError("The distribution count must equal the sum "
                                 "of bucket counts")
        self._count = count
        self._sum = sum_
        self._sum_of_squared_deviation = sum_of_squared_deviation
        self._bucket_options = bucket_options
        self._buckets = buckets

    def __repr__(self):
        try:
            bounds = self.bucket_options.type_.bounds,
        except AttributeError:
            bounds = None

        return ("{}({})"
                .format(
                    type(self).__name__,
                    bounds
                ))

    @property
    def count(self):
        return self._count

    @property
    def sum(self):
        return self._sum

    @property
    def sum_of_squared_deviation(self):
        return self._sum_of_squared_deviation

    @property
    def bucket_options(self):
        return self._bucket_options

    @property
    def buckets(self):
        return self._buckets



# Copyright 2018, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from opencensus.trace.tracestate import _KEY_FORMAT, _VALUE_FORMAT, Tracestate

_DELIMITER_FORMAT = '[ \t]*,[ \t]*'
_MEMBER_FORMAT = '(%s)(=)(%s)' % (_KEY_FORMAT, _VALUE_FORMAT)

_DELIMITER_FORMAT_RE = re.compile(_DELIMITER_FORMAT)
_MEMBER_FORMAT_RE = re.compile(_MEMBER_FORMAT)


class TracestateStringFormatter(object):
    def from_string(self, string):
        tracestate = Tracestate()
        for member in re.split(_DELIMITER_FORMAT_RE, string):
            match = _MEMBER_FORMAT_RE.match(member)
            if not match:
                raise ValueError('illegal key-value format %r' % (member))
            key, eq, value = match.groups()
            if key in tracestate:
                raise ValueError('conflict key {!r}'.format(key))
            tracestate[key] = value
        return tracestate

    def to_string(self, tracestate):
        return ','.join(map(
            lambda key: key + '=' + tracestate[key],
            tracestate
        ))



# Copyright 2016-17, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing base class for exporters."""


class Exporter(object):
    """Base class for opencensus trace request exporters.

    Subclasses of :class:`Exporter` must override :meth:`export`.
    """

    def emit(self, span_datas):
        """
        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to emit
        """
        raise NotImplementedError

    def export(self, span_datas):
        """Export the trace. Send trace to transport, and transport will call
        exporter.emit() to actually send the trace to the specified tracing
        backend.

        :type span_datas: list of :class:
            `~opencensus.trace.span_data.SpanData`
        :param list of opencensus.trace.span_data.SpanData span_datas:
            SpanData tuples to export
        """
        raise NotImplementedError
